from typing import Text, Optional
from .link import Link

try:
    from navec import Navec
    from slovnet import NER

    import_error = False
except ImportError:
    import_error = True

DEFAULT_MAPPING = {
    "PER": "<person>",
    "LOC": "<location>",
    "ORG": "<organization>",
}


class Ner(Link):
    """
    Replaces named entities with specified tags using NLP.
    """

    def __init__(
        self,
        navec_path: str,
        ner_path: str,
        mapping: dict[str, str] = DEFAULT_MAPPING,
        level=0,
    ):
        super().__init__(level)
        self.navec_path = navec_path
        self.ner_path = ner_path
        self.mapping = mapping

        if import_error:
            raise RuntimeError('requires "ner" feature enabled')

        try:
            self.navec = Navec.load(navec_path)
        except ValueError:
            raise RuntimeError(
                "download navec model from repository: https://github.com/natasha/navec#downloads"
            )

        try:
            self.ner = NER.load(ner_path)
        except ValueError:
            raise RuntimeError(
                "download slovnet model from repository: https://github.com/natasha/slovnet#downloads"
            )

        self.ner.navec(self.navec)

    def process(self, text: Text):
        ner = self.ner(text)
        for span in ner.spans:
            if span.type in self.mapping:
                span_text = ner.text[span.start : span.stop]
                text = text.replace(span_text, self.mapping[span.type])
        return text

    def __repr__(self):
        return self._indentation + f'{type(self).__name__}("{self.model_name}")'



from typing import Sequence, Text
from .link import Link


class Chain(Link):
    """
    Executes text processing through a sequence of chained transformation steps.
    """

    def __init__(self, chain: Sequence[Link], level=0):
        super().__init__(level)
        self.chain = []
        for link in chain:
            link._level(level + 1)
            self.chain.append(link)

    def process(self, text: Text):
        for link in self.chain:
            text = link.process(text)
        return text

    def __repr__(self):
        output = self._indentation + "Chain ["
        for link in self.chain:
            output += "\n" + self._indentation + str(link) + ","
        output += "\n" + self._indentation + "]"
        return output



from typing import Text
import unicodedata
from .normalize_unicode import NormalizeUnicode
from .chain import Chain
from .replace import Replace

MINUS = "-"
DOUBLE_MINUS = 2 * MINUS

APOSTROPHE = "'"
QUOTE = '"'
ELLIPSIS = "..."

normalize = Chain(
    [
        # Hyphens
        Replace("\u002d", MINUS),
        Replace("\u00ad", MINUS),
        Replace("\u2010", MINUS),
        Replace("\u2011", MINUS),
        Replace("\u2012", MINUS),

        # En/Em dashes
        Replace("\u2013", MINUS),
        Replace("\u2014", DOUBLE_MINUS),
        Replace("\u2015", DOUBLE_MINUS),
        Replace("\u2e3a", DOUBLE_MINUS),
        Replace("\u2e3b", DOUBLE_MINUS),

        # Mathematical minus
        Replace("\u2212", MINUS),

        # Other
        Replace("\u058a", MINUS),
        Replace("\u1806", MINUS),

        # Single quotes
        Replace("\u2018", APOSTROPHE),
        Replace("\u2019", APOSTROPHE),
        Replace("\u201a", APOSTROPHE),
        Replace("\u201b", APOSTROPHE),

        # Double quotes
        Replace("\u201c", QUOTE),
        Replace("\u201d", QUOTE),
        Replace("\u201e", QUOTE),
        Replace("\u201f", QUOTE),

        # Angle quotes
        Replace("\u00ab", QUOTE),
        Replace("\u00bb", QUOTE),
        Replace("\u2039", QUOTE),
        Replace("\u203a", QUOTE),

        # Prime quotes
        Replace("\u2032", APOSTROPHE),
        Replace("\u2033", QUOTE),
        Replace("\u2035", APOSTROPHE),
        Replace("\u2036", QUOTE),

        # CJK quotes
        Replace("\u300c", QUOTE),
        Replace("\u300d", QUOTE),
        Replace("\u301d", QUOTE),
        Replace("\u301e", QUOTE),

        # Ellipsis
        Replace("\u2026", ELLIPSIS),
        Replace("\u22ef", ELLIPSIS),
        Replace("\u00b7", MINUS),

        # Fractions
        Replace("\u00bc", "1/4"),
        Replace("\u00bd", "1/2"),
        Replace("\u00be", "3/4"),

        # Arrows
        Replace("\u2192", "->"),
        Replace("\u2190", "<-"),
        Replace("\u2194", "<->"),
    ]
)

class Asciify(NormalizeUnicode):
    """
    Converts Unicode text to ASCII by removing diacritics and non-ASCII characters.
    """

    def __init__(self, level=0):
        super().__init__("NFKD", level)

    def process(self, text: Text):
        text = super().process(text)
        text = normalize(text)
        text = "".join(
            [
                character
                for character in text
                if not unicodedata.combining(character)
            ]
        )
        return text

    def __repr__(self):
        return self._indentation + type(self).__name__



from .chain import Chain
from .collapse_whitespace import CollapseWhitespace
from .tab_to_whitespace import TabToWhitespace
from .lowercase import Lowercase
from .uppercase import Uppercase
from .trim import Trim
from .replace import Replace
from .pattern_replace import PatternReplace
from .normalize_unicode import NormalizeUnicode
from .asciify import Asciify
from .ftfy import Ftfy
from .translate import Translate
from .gpt import Gpt
from .remove_punctuation import RemovePunctuation
from .ner import Ner
from .anonymize import Anonymize



from typing import Text
from .link import Link
import re


class PatternReplace(Link):
    """
    Replaces text matching a regex pattern with specified replacement.
    """

    def __init__(self, pattern: Text, to: Text):
        self.raw_pattern = pattern
        self.pattern = re.compile(pattern)
        self.to = to

    def process(self, text: Text):
        return self.pattern.sub(self.to, text)

    def __repr__(self):
        return (
            self._indentation
            + f'{type(self).__name__}("{self.raw_pattern}" -> "{self.to}")'
        )



from typing import Text
from .link import Link

try:
    import g4f

    g4f.debug.logging = False
    g4f.debug.version_check = False
    import_error = False
except ImportError:
    import_error = True


class Gpt(Link):
    """
    Processes text using AI.
    """

    def __init__(self, model: Text, prompt: Text, level=0):
        super().__init__(level)
        self.model = model
        self.prompt = prompt

    def process(self, text: Text):
        if import_error:
            raise RuntimeError('requires "gpt" feature enabled')

        return g4f.ChatCompletion.create(
            self.model,
            [
                {"role": "system", "content": self.prompt},
                {"role": "user", "content": text},
            ],
        )

    def __repr__(self):
        return (
            self._indentation
            + f'{type(self).__name__}("{self.model}", "{self.prompt}")'
        )



from typing import Text
import unicodedata
from .link import Link

DEFAULT_FORM = "NFD"


class NormalizeUnicode(Link):
    """
    Normalizes Unicode text to a specified form.
    """

    def __init__(self, form=DEFAULT_FORM, level=0):
        super().__init__(level)
        self.form = form

    def process(self, text: Text):
        return unicodedata.normalize(self.form, text)

    def __repr__(self):
        return self._indentation + f'{type(self).__name__}("{self.form}")'



from typing import Text
from .link import Link

try:
    import ftfy

    import_error = False
except ImportError:
    import_error = True


class Ftfy(Link):
    """
    Fixes text encoding issues and mojibake using ftfy library corrections.
    """

    def process(self, text: Text):
        if import_error:
            raise RuntimeError('requires "ftfy" feature enabled')

        return ftfy.fix_text(text, normalization=None)



from .pattern_replace import PatternReplace
from .chain import Chain


DEFAULT_MAPPING = {
    "<email>": "<email>",
    "<phone>": "<phone>",
    "<snils>": "<snils>",
    "<passport>": "<passport>",
    "<inn>": "<inn>",
    "<ogrn>": "<ogrn>",
    "<vu>": "<vu>",
    "<card>": "<card>",
}


class Anonymize(Chain):
    """
    Collapses multiple consecutive spaces into a single space.

    # Output Masking

    - `<email>` -- Email address
    - `<phone>` -- Phone number
    - `<snils>` -- SNILS code
    - `<passport>` -- Passport code
    - `<inn>` -- INN code
    - `<ogrn>` -- Company registration number
    - `<vu>` -- Driver license
    - `<card>` -- Bank card code
    """

    def __init__(self, mapping: dict[str, str] = DEFAULT_MAPPING, level=0):
        self.mapping = mapping
        M = mapping
        chain = [
            PatternReplace(
                r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}\b", M["<email>"]
            ),
            PatternReplace(
                r"\+?(7|8)[\s\-]?\(?\d{3}\)?[\s\-]?\d{3}[\s\-]?\d{2}[\s\-]?\d{2}",
                M["<phone>"],
            ),
            PatternReplace(r"\d{3}-\d{3}-\d{3}\s\d{2}", M["<snils>"]),
            PatternReplace(r"\b\d{2}\s?\d{2}\s?\d{6}\b", M["<passport>"]),
            PatternReplace(r"\b\d{10}\b|\b\d{12}\b", M["<inn>"]),
            PatternReplace(r"\b\d{13,15}\b", M["<ogrn>"]),
            PatternReplace(r"\b\d{2}[- ]?\d{2}[- ]?\d{6}\b", M["<vu>"]),
            PatternReplace(r"\b\d{4}[- ]?\d{4}[- ]?\d{4}[- ]?\d{4}\b", M["<card>"]),
        ]
        super().__init__(chain, level=0)

    def __repr__(self):
        return self._indentation + f"{type(self).__name__}"



from typing import Text
from .link import Link


class Trim(Link):
    """
    Removes leading and trailing whitespace from text.
    """

    def process(self, text: Text):
        return text.strip()



from typing import Text
from .link import Link
import asyncio

try:
    import googletrans
except ImportError:
    googletrans = None


class Translate(Link):
    """
    Translates text to a specified language using the Google translation service.
    """

    def __init__(self, language: Text, level=0):
        super().__init__(level)
        self.language = language

        if googletrans is None:
            raise RuntimeError('requires "translate" feature enabled')

        self.service = googletrans.Translator()

    def process(self, text: Text):
        async def translate():
            return await self.service.translate(text, dest=self.language)

        return asyncio.run(translate()).text

    def __repr__(self):
        return self._indentation + f'{type(self).__name__}("{self.language}")'



from typing import Text
from .link import Link
import unicodedata


class RemovePunctuation(Link):
    """
    Removes all punctuation characters.
    """

    def is_punctuation(self, character: str):
        category = unicodedata.category(character)
        return category.startswith("P")

    def process(self, text: Text):
        return "".join(
            [character for character in text if not self.is_punctuation(character)]
        )



from typing import Text
from .link import Link


class Replace(Link):
    """
    Replaces all occurrences of a specified substring with another substring.
    """

    def __init__(self, what: Text, to: Text):
        self.what = what
        self.to = to

    def process(self, text: Text):
        return text.replace(self.what, self.to)

    def __repr__(self):
        return (
            self._indentation + f'{type(self).__name__}("{self.what}" -> "{self.to}")'
        )



from .pattern_replace import PatternReplace


class CollapseWhitespace(PatternReplace):
    """
    Collapses multiple consecutive spaces into a single space.
    """

    def __init__(self):
        super().__init__(" {2,}", " ")

    def __repr__(self):
        return self._indentation + type(self).__name__



from typing import Text


class Link:
    """
    Base class for creating text processing steps in a transformation chain.
    """

    def __init__(self, level=0):
        self._level(level)

    def _level(self, level):
        self._indentation = " " * 4 * level

    def process(self, text: Text):
        raise NotImplementedError(
            'implement your text-processing logic in the "process" function'
        )

    def __repr__(self):
        return self._indentation + type(self).__name__

    def __call__(self, text: Text):
        return self.process(text)



from .replace import Replace

DEFAULT_LENGTH = 4


class TabToWhitespace(Replace):
    """
    Replaces tab characters with equivalent whitespace spaces.
    """

    def __init__(self, length=DEFAULT_LENGTH):
        self.length = length
        super().__init__("\t", " " * self.length)

    def __repr__(self):
        return self._indentation + f"{type(self).__name__}({self.length})"
